@@ -88,32 +88,42 @@ aggregate.sf = function(x, by, FUN, ..., do_union = TRUE, simplify = TRUE,
 	}
 }
 
-
 #' Areal-weighted interpolation of polygon data
 #'
 #' Areal-weighted interpolation of polygon data
 #' @param x object of class \code{sf}, for which we want to aggregate attributes
 #' @param to object of class \code{sf} or \code{sfc}, with the target geometries
 #' @param extensive logical; if TRUE, the attribute variables are assumed to be spatially extensive (like population) and the sum is preserved, otherwise, spatially intensive (like population density) and the mean is preserved.
+#' @param ... ignored
 #' @examples
 #' nc = st_read(system.file("shape/nc.shp", package="sf"))
 #' g = st_make_grid(nc, n = c(20,10))
 #' a1 = st_interpolate_aw(nc["BIR74"], g, extensive = FALSE)
 #' sum(a1$BIR74) / sum(nc$BIR74) # not close to one: property is assumed spatially intensive
 #' a2 = st_interpolate_aw(nc["BIR74"], g, extensive = TRUE)
+#' # verify mass preservation (pycnophylactic) property:
 #' sum(a2$BIR74) / sum(nc$BIR74)
 #' a1$intensive = a1$BIR74
 #' a1$extensive = a2$BIR74
-#' plot(a1[c("intensive", "extensive")])
+#' plot(a1[c("intensive", "extensive")], key.pos = 4)
 #' @export
-st_interpolate_aw = function(x, to, extensive) {
+st_interpolate_aw = function(x, to, extensive, ...) UseMethod("st_interpolate_aw")
+
+#' @export
+st_interpolate_aw.sf = function(x, to, extensive, ...) {
 	if (!inherits(to, "sf") && !inherits(to, "sfc"))
 		stop("st_interpolate_aw requires geometries in argument to")
 	if (! all_constant(x))
 		warning("st_interpolate_aw assumes attributes are constant over areas of x")
 	i = st_intersection(st_geometry(x), st_geometry(to))
 	idx = attr(i, "idx")
-	i = st_cast(i, "MULTIPOLYGON")
+
+	# https://stackoverflow.com/questions/57767022/how-do-you-use-st-interpolate-aw-with-polygon-layers-that-legitimately-include-p
+	gc = which(st_is(i, "GEOMETRYCOLLECTION"))
+	i[gc] = st_collection_extract(i[gc], "POLYGON")
+	two_d = which(st_dimension(i) == 2)
+	i[two_d] = st_cast(i[two_d], "MULTIPOLYGON")
+
 	x$...area_s = unclass(st_area(x))
 	st_geometry(x) = NULL # sets back to data.frame
 	x = x[idx[,1], ]      # create st table
@@ -130,6 +140,5 @@ st_interpolate_aw = function(x, to, extensive) {
 	x = aggregate(x, list(idx[,2]), sum)
 	df = st_sf(x, geometry = st_geometry(to)[x$Group.1])
 	df$...area_t = df$...area_st = df$...area_s = NULL
-	st_agr(df) = "aggregate"
-	df
+	st_set_agr(df, "aggregate")
 }
