@@ -6,7 +6,7 @@ is.na.bbox = function(x) identical(x, NA_bbox_)
 
 bb_wrap = function(bb) {
 	stopifnot(is.numeric(bb) && length(bb) == 4)
-	structure(bb, names = c("xmin", "ymin", "xmax", "ymax"), class = "bbox")
+	structure(as.double(bb), names = c("xmin", "ymin", "xmax", "ymax"), class = "bbox")
 }
 
 bbox.Set = function(obj, ...) {
