@@ -98,13 +98,15 @@ plot.sf <- function(x, y, ..., main, pal = NULL, nbreaks = 10, breaks = "pretty"
 			cols = cols[1:max.plot]
 
 		if (!is.null(key.pos)) {
-			values = as.numeric(as.matrix(as.data.frame(x)[cols])) # will kill factors, and also units
+			values = do.call(c, as.data.frame(x)[cols])
 			if (logz)
 				values = log10(values)
 			if (is.character(breaks)) { # compute breaks from values:
-				n.unq = length(unique(na.omit(values)))
+				v0 = values[!is.na(values)]
+				n.unq = length(unique(v0))
 				breaks = if (! all(is.na(values)) && n.unq > 1)
-						classInt::classIntervals(na.omit(values), min(nbreaks, n.unq), breaks, warnSmallN = FALSE)$brks
+						classInt::classIntervals(v0, min(nbreaks, n.unq), 
+							breaks, warnSmallN = FALSE)$brks
 					else
 						range(values, na.rm = TRUE) # lowest and highest!
 				nbreaks = length(breaks) - 1
@@ -151,9 +153,6 @@ plot.sf <- function(x, y, ..., main, pal = NULL, nbreaks = 10, breaks = "pretty"
 			if (is.list(values))
 				stop("plotting list-columns not supported") # nocov
 
-			if (inherits(values, "POSIXt"))
-				values = as.numeric(values)
-
 			if (is.character(values))
 				values = as.factor(values)
 			else if (logz)
@@ -174,14 +173,17 @@ plot.sf <- function(x, y, ..., main, pal = NULL, nbreaks = 10, breaks = "pretty"
 								pal
 						colors[as.numeric(values)]
 					} else {
-						values = as.numeric(values) # drop units, if any
+						if (! inherits(values, c("POSIXt", "Date")))
+							values = as.numeric(values) # drop units, if any
 						if (is.character(breaks)) { # compute breaks from values:
-							n.unq = length(unique(na.omit(values)))
+							v0 = values[!is.na(values)]
+							n.unq = length(unique(v0))
 							breaks = if (! all(is.na(values)) && n.unq > 1)
-								classInt::classIntervals(na.omit(values), min(nbreaks, n.unq), breaks, warnSmallN = FALSE)$brks
+								classInt::classIntervals(v0, min(nbreaks, n.unq),
+									breaks, warnSmallN = FALSE)$brks
 							else
 								range(values, na.rm = TRUE) # lowest and highest!
-						} 
+						}
 						# this is necessary if breaks were specified either as character or as numeric
 						# "pretty" takes nbreaks as advice only:
 						nbreaks = length(breaks) - 1
@@ -189,7 +191,7 @@ plot.sf <- function(x, y, ..., main, pal = NULL, nbreaks = 10, breaks = "pretty"
 						cuts = if (all(is.na(values)))
 								rep(NA_integer_, length(values))
 							else if (diff(range(values, na.rm = TRUE)) == 0)
-								values * 0 + 1  # preserves NA's
+								ifelse(is.na(values), NA_integer_, 1L)
 							else
 								cut(values, breaks, include.lowest = TRUE)
 						colors = if (is.function(pal))
@@ -250,8 +252,8 @@ plot.sf <- function(x, y, ..., main, pal = NULL, nbreaks = 10, breaks = "pretty"
 			localTitle(main, ...)
 		}
 	}
-	if (reset) {
-		layout(matrix(1)) # reset
+	if (!isTRUE(dots$add) && reset) { # reset device:
+		layout(matrix(1))
 		desel = which(names(opar) %in% c("cin", "cra", "csi", "cxy", "din", "page", "fig"))
 		par(opar[-desel])
 	}
@@ -291,9 +293,9 @@ plot.sfc_POINT = function(x, y, ..., pch = 1, cex = 1, col = 1, bg = 0, lwd = 1,
 	col = rep(col, length.out = npts)
 	bg = rep(bg, length.out = npts)
 	cex = rep(cex, length.out = npts)
-	mat = do.call(rbind, x)
+	mat = t(matrix(unlist(x, use.names = FALSE), ncol = length(x))) #933
 	if (!is.null(mat)) {
-		ne = apply(mat, 1, function(x) all(is.finite(x))) # ne: not empty
+		ne = !is.na(rowMeans(mat))  ## faster than apply; #933
 		points(mat[ne,, drop = FALSE], pch = pch[ne], col = col[ne], bg = bg[ne],
 			cex = cex[ne], lwd = lwd, lty = lty, type = type)
 	}
@@ -478,7 +480,7 @@ plot.sfc_GEOMETRYCOLLECTION = function(x, y, ..., pch = 1, cex = 1, bg = 0, lty
 #' @method plot sfc_GEOMETRY
 #' @export
 plot.sfc_GEOMETRY = function(x, y, ..., pch = 1, cex = 1, bg = 0, lty = 1, lwd = 1,
-	col = 1, border = 1, add = FALSE) {
+	col = ifelse(st_dimension(x) == 2, NA, 1), border = 1, add = FALSE) {
 	stopifnot(missing(y))
 	if (! add)
 		plot_sf(x, ...)
@@ -526,13 +528,13 @@ plot.sfg = function(x, ...) {
 #'
 #' The default aspect for map plots is 1; if however data are not
 #' projected (coordinates are long/lat), the aspect is by default set to
-#' 1/cos(My * pi)/180) with My the y coordinate of the middle of the map
+#' 1/cos(My * pi/180) with My the y coordinate of the middle of the map
 #' (the mean of \code{ylim}, which defaults to the y range of bounding box). This
 #' implies an \href{https://en.wikipedia.org/wiki/Equirectangular_projection}{Equirectangular projection}.
 #'
 plot_sf = function(x, xlim = NULL, ylim = NULL, asp = NA, axes = FALSE, bgc = par("bg"), ...,
     xaxs, yaxs, lab, setParUsrBB = FALSE, bgMap = NULL, expandBB = c(0,0,0,0), graticule = NA_crs_,
-	col_graticule = 'grey') {
+	col_graticule = 'grey', border) {
 
 # sp's bbox: matrix
 #   min max
@@ -793,9 +795,9 @@ bb2merc = function(x, cls = "ggmap") { # return bbox in the appropriate "web mer
 		c(offset,offset,0,0)) 
 	for(i in seq_along(poly)) {
 		if (key.pos %in% c(1,3))
-			polygon(poly[[i]], c(0, 0, 1, 1) + offs, col=col[i], border=NA)
+			polygon(poly[[i]], c(0, 0, 1, 1) + offs, col = col[i], border = NA)
 		if (key.pos %in% c(2,4))
-			polygon(c(0, 0, 1, 1) + offs, poly[[i]], col=col[i], border=NA)
+			polygon(c(0, 0, 1, 1) + offs, poly[[i]], col = col[i], border = NA)
 	}
 
 	# box() now would draw around [0,1]:
@@ -807,7 +809,9 @@ bb2merc = function(x, cls = "ggmap") { # return bbox in the appropriate "web mer
 
 	labels = if (logz)
 			parse(text = paste0("10^", at))
-		else 
+		else if (inherits(breaks, c("POSIXt", "Date")))
+			format(at)
+		else
 			TRUE
 
 	if (add.axis)
