@@ -61,17 +61,36 @@ set_utf8 = function(x) {
 #' without warnings. \code{promote_to_multi} is handled on a per-geometry column
 #' basis; \code{type} may be specified for each geometry column.
 #'
+#' Note that stray files in data source directories (such as \code{*.dbf}) may
+#' lead to spurious errors that accompanying \code{*.shp} are missing.
+#'
 #' In case of problems reading shapefiles from USB drives on OSX, please see
 #' \url{https://github.com/r-spatial/sf/issues/252}.
+#'
+#' For \code{query} with a character \code{dsn} the query text is handed to
+#' 'ExecuteSQL' on the GDAL/OGR data set and will result in the creation of a
+#' new layer (and \code{layer} is ignored). See 'OGRSQL'
+#' \url{https://gdal.org/user/ogr_sql_dialect.html} for details. Please note that the
+#' 'FID' special field is driver-dependent, and may be either 0-based (e.g. ESRI
+#' Shapefile), 1-based (e.g. MapInfo) or arbitrary (e.g. OSM). Other features of
+#' OGRSQL are also likely to be driver dependent. The available layer names may
+#' be obtained with
+#' \link{st_layers}. Care will be required to properly escape the use of some layer names.
+#'
 #' @return object of class \link{sf} when a layer was successfully read; in case
 #'   argument \code{layer} is missing and data source \code{dsn} does not
 #'   contain a single layer, an object of class \code{sf_layers} is returned
 #'   with the layer names, each with their geometry type(s). Note that the
 #'   number of layers may also be zero.
+#' @seealso \link{st_layers}, \link{st_drivers}
 #' @examples
 #' nc = st_read(system.file("shape/nc.shp", package="sf"))
 #' summary(nc) # note that AREA was computed using Euclidian area on lon/lat degrees
 #'
+#' ## only three fields by select clause
+#' ## only two features by where clause
+#' nc_sql = st_read(system.file("shape/nc.shp", package="sf"),
+#'                      query = "SELECT NAME, SID74, FIPS FROM \"nc\" WHERE BIR74 > 20000")
 #' \dontrun{
 #'   library(sp)
 #'   example(meuse, ask = FALSE, echo = FALSE)
@@ -81,42 +100,51 @@ set_utf8 = function(x) {
 #'   if (exists("st_meuse"))
 #'     summary(st_meuse)
 #' }
+#'
+#' \dontrun{
+#' ## note that we need special escaping of layer  within single quotes (nc.gpkg)
+#' ## and that geom needs to be included in the select, otherwise we don't detect it
+#' layer <- st_layers(system.file("gpkg/nc.gpkg", package = "sf"))$name[1]
+#' nc_gpkg_sql = st_read(system.file("gpkg/nc.gpkg", package = "sf"),
+#'    query = sprintf("SELECT NAME, SID74, FIPS, geom  FROM \"%s\" WHERE BIR74 > 20000", layer))
+#' }
 #' @export
 st_read = function(dsn, layer, ...) UseMethod("st_read")
 
-#' @name st_read
-#' @note The use of \code{system.file} in examples make sure that examples run regardless where R is installed:
-#' typical users will not use \code{system.file} but give the file name directly, either with full path or relative
-#' to the current working directory (see \link{getwd}). "Shapefiles" consist of several files with the same basename
-#' that reside in the same directory, only one of them having extension \code{.shp}.
 #' @export
-st_read.default = function(dsn, layer, ..., options = NULL, quiet = FALSE, geometry_column = 1L, type = 0,
-		promote_to_multi = TRUE, stringsAsFactors = default.stringsAsFactors(),
-		int64_as_string = FALSE, check_ring_dir = FALSE) {
-
+st_read.default = function(dsn, layer, ...) {
 	if (missing(dsn))
 		stop("dsn should specify a data source or filename")
+	else {
+		dsn_is_null = is.null(dsn)
+		class_dsn = class(dsn)
+		tr <- try(dsn <- as.character(dsn))
+		if (dsn_is_null || inherits(tr, "try-error"))
+			stop(paste("no st_read method available for objects of class", paste(class_dsn, collapse = ", ")))
+		else
+			st_read.character(dsn, layer, ...)
+	}
+}
 
-	layer = if (missing(layer))
-		character(0)
-	else
-		enc2utf8(layer)
-
-	if (length(dsn) == 1 && file.exists(dsn))
-		dsn = enc2utf8(normalizePath(dsn))
-
-	if (length(promote_to_multi) > 1)
-		stop("`promote_to_multi' should have length one, and applies to all geometry columns")
-
-	x = CPL_read_ogr(dsn, layer, as.character(options), quiet, type, promote_to_multi, int64_as_string)
+process_cpl_read_ogr = function(x, quiet = FALSE, ..., check_ring_dir = FALSE,
+		stringsAsFactors = ifelse(as_tibble, FALSE, default.stringsAsFactors()), geometry_column = 1, as_tibble = FALSE) {
 
 	which.geom = which(vapply(x, function(f) inherits(f, "sfc"), TRUE))
 
+	if (as_tibble && !requireNamespace("tibble", quietly = TRUE))
+		stop("package tibble not available: install first?")
+
 	# in case no geometry is present:
 	if (length(which.geom) == 0) {
 		warning("no simple feature geometries present: returning a data.frame or tbl_df",
 			call. = FALSE)
-		return(as.data.frame(x , stringsAsFactors = stringsAsFactors))
+		x = if (!as_tibble) {
+				if (any(sapply(x, is.list)))
+					warning("list-column(s) present: in case of failure, try read_sf or as_tibble=TRUE") # nocov
+				as.data.frame(x , stringsAsFactors = stringsAsFactors)
+			} else
+				tibble::as_tibble(x)
+		return(x)
 	}
 
 	nm = names(x)[which.geom]
@@ -127,10 +155,18 @@ st_read.default = function(dsn, layer, ..., options = NULL, quiet = FALSE, geome
 	list.cols = x[lc.other]
 	nm.lc = names(x)[lc.other]
 
-	x = if (length(x) == length(geom)) # ONLY geometry column(s)
-		data.frame(row.names = seq_along(geom[[1]]))
-	else
-		as.data.frame(set_utf8(x[-c(lc.other, which.geom)]), stringsAsFactors = stringsAsFactors)
+	if (length(x) == length(geom)) { # ONLY geometry column(s)
+		if (as_tibble)
+			x <- tibble::tibble(row.names = seq_along(geom[[1]]))[-1]
+		else
+			x <- data.frame(row.names = seq_along(geom[[1]]))
+	} else {
+		x <- as.data.frame(set_utf8(x[-c(lc.other, which.geom)]), stringsAsFactors = stringsAsFactors)
+		if (as_tibble) {
+			# "sf" class is added later by `st_as_sf` (and sets all the attributes)
+			x <- tibble::new_tibble(x, attributes(x), nrow = nrow(x))
+		}
+	}
 
 	for (i in seq_along(lc.other))
 		x[[ nm.lc[i] ]] = list.cols[[i]]
@@ -147,6 +183,40 @@ st_read.default = function(dsn, layer, ..., options = NULL, quiet = FALSE, geome
 		x
 }
 
+#' @name st_read
+#' @param fid_column_name character; name of column to write feature IDs to; defaults to not doing this
+#' @param drivers character; limited set of driver short names to be tried (default: try all)
+#' @note The use of \code{system.file} in examples make sure that examples run regardless where R is installed:
+#' typical users will not use \code{system.file} but give the file name directly, either with full path or relative
+#' to the current working directory (see \link{getwd}). "Shapefiles" consist of several files with the same basename
+#' that reside in the same directory, only one of them having extension \code{.shp}.
+#' @export
+st_read.character = function(dsn, layer, ..., query = NA, options = NULL, quiet = FALSE, geometry_column = 1L, type = 0,
+		promote_to_multi = TRUE, stringsAsFactors = default.stringsAsFactors(),
+		int64_as_string = FALSE, check_ring_dir = FALSE, fid_column_name = character(0),
+		drivers = character(0)) {
+
+	layer = if (missing(layer))
+		character(0)
+	else
+		enc2utf8(layer)
+	if (nchar(dsn) < 1) {
+		stop("`dsn` must point to a source, not an empty string.", call. = FALSE)
+	}
+	dsn_exists = file.exists(dsn)
+	dsn_isdb = is_db_driver(dsn)
+	if (length(dsn) == 1 && dsn_exists && !dsn_isdb)
+		dsn = enc2utf8(normalizePath(dsn))
+
+	if (length(promote_to_multi) > 1)
+		stop("`promote_to_multi' should have length one, and applies to all geometry columns")
+
+	x = CPL_read_ogr(dsn, layer, query, as.character(options), quiet, type, fid_column_name,
+		drivers, promote_to_multi, int64_as_string, dsn_exists, dsn_isdb)
+	process_cpl_read_ogr(x, quiet, check_ring_dir = check_ring_dir,
+		stringsAsFactors = stringsAsFactors, geometry_column = geometry_column, ...)
+}
+
 #' @name st_read
 #' @export
 #' @details \code{read_sf} and \code{write_sf} are aliases for \code{st_read} and \code{st_write}, respectively, with some
@@ -160,16 +230,8 @@ st_read.default = function(dsn, layer, ..., options = NULL, quiet = FALSE, geome
 #'    "[[3.2,4],[3,4.6],[3.8,4.4],[3.5,3.8],[3.4,3.6],[3.9,4.5]]}")
 #' x = read_sf(geojson_txt)
 #' x
-read_sf <- function(..., quiet = TRUE, stringsAsFactors = FALSE) {
-	if (! requireNamespace("tibble", quietly = TRUE))
-		stop("package tibble not available: install first?")
-	tbl = tibble::as_tibble(as.data.frame(
-		st_read(..., quiet = quiet, stringsAsFactors = stringsAsFactors)))
-	has_geom = any(vapply(tbl, function(f) inherits(f, "sfc"), TRUE))
-	if (has_geom) 
-		st_as_sf(tbl)
-	else
-		tbl
+read_sf <- function(..., quiet = TRUE, stringsAsFactors = FALSE, as_tibble = TRUE) {
+	st_read(..., quiet = quiet, stringsAsFactors = stringsAsFactors, as_tibble = as_tibble)
 }
 
 clean_columns = function(obj, factorsAsCharacter) {
@@ -237,29 +299,49 @@ abbreviate_shapefile_names = function(x) {
 #' are found at \url{http://www.gdal.org/ogr_formats.html}.
 #' @param ... other arguments passed to \link{dbWriteTable} when \code{dsn} is a
 #' Database Connection
-#' @param dataset_options character; driver dependent dataset creation options; multiple options supported.
-#' @param layer_options character; driver dependent layer creation options; multiple options supported.
+#' @param dataset_options character; driver dependent dataset creation options; 
+#' multiple options supported.
+#' @param layer_options character; driver dependent layer creation options; 
+#' multiple options supported.
 #' @param quiet logical; suppress info on name, driver, size and spatial reference
-#' @param factorsAsCharacter logical; convert \code{factor} objects into character strings (default), else into numbers by
-#' \code{as.numeric}.
-#' @param update logical; \code{FALSE} by default for single-layer drivers but \code{TRUE} by default for database drivers
-#' as defined by \code{db_drivers}.
-#' For database-type drivers (e.g. GPKG) \code{TRUE} values will make \code{GDAL} try
-#' to update (append to) the existing data source, e.g. adding a table to an existing database.
-#' @param delete_dsn logical; delete data source \code{dsn} before attempting to write?
-#' @param delete_layer logical; delete layer \code{layer} before attempting to write? (not yet implemented)
-#' @details columns (variables) of a class not supported are dropped with a warning. When deleting layers or
-#' data sources is not successful, no error is emitted. \code{delete_dsn} and \code{delete_layers} should be
+#' @param factorsAsCharacter logical; convert \code{factor} objects into 
+#' character strings (default), else into numbers by \code{as.numeric}.
+#' @param update logical; \code{FALSE} by default for single-layer drivers but 
+#' \code{TRUE} by default for database drivers as defined by \code{db_drivers}.
+#' For database-type drivers (e.g. GPKG) \code{TRUE} values will make 
+#' \code{GDAL} try to update (append to) the existing data source, e.g. adding
+#' a table to an existing database, or adding records to a layer. See also the
+#' next two arguments and Details.
+#' @param delete_dsn logical; delete data source \code{dsn} before attempting 
+#' to write?
+#' @param delete_layer logical; delete layer \code{layer} before attempting to
+#' write?
+#' @param fid_column_name character, name of column with feature IDs; if
+#' specified, this column is no longer written as feature attribute.
+#' @details 
+#' Columns (variables) of a class not supported are dropped with a warning. 
+#' 
+#' When updating an existing layer, records are appended to it if the updating
+#' object has the right variable names and types. If names don't match an 
+#' error is raised. If types don't match, behaviour is undefined: GDAL may
+#' raise warnings or errors or fail silently.
+#' 
+#' When deleting layers or data sources is not successful, no error is emitted. 
+#' \code{delete_dsn} and \code{delete_layers} should be
 #' handled with care; the former may erase complete directories or databases.
 #' @seealso \link{st_drivers}
+#' @return \code{obj}, invisibly; in case \code{obj} is of class \code{sfc}, 
+#' it is returned as an \code{sf} object.
 #' @examples
 #' nc = st_read(system.file("shape/nc.shp", package="sf"))
-#' st_write(nc, "nc.shp")
-#' st_write(nc, "nc.shp", delete_layer = TRUE) # overwrites
+#' st_write(nc, paste0(tempdir(), "/", "nc.shp"))
+#' st_write(nc, paste0(tempdir(), "/", "nc.shp"), delete_layer = TRUE) # overwrites
 #' data(meuse, package = "sp") # loads data.frame from sp
 #' meuse_sf = st_as_sf(meuse, coords = c("x", "y"), crs = 28992)
-#' st_write(meuse_sf, "meuse.csv", layer_options = "GEOMETRY=AS_XY") # writes X and Y as columns
-#' st_write(meuse_sf, "meuse.csv", layer_options = "GEOMETRY=AS_WKT", delete_dsn=TRUE) # overwrites
+#' # writes X and Y as columns:
+#' st_write(meuse_sf, paste0(tempdir(), "/", "meuse.csv"), layer_options = "GEOMETRY=AS_XY") 
+#' st_write(meuse_sf, paste0(tempdir(), "/", "meuse.csv"), layer_options = "GEOMETRY=AS_WKT",
+#'   delete_dsn=TRUE) # overwrites
 #' \dontrun{
 #'  library(sp)
 #'  example(meuse, ask = FALSE, echo = FALSE)
@@ -285,21 +367,37 @@ st_write.sfc = function(obj, dsn, layer, ...) {
 st_write.sf = function(obj, dsn, layer = NULL, ...,
 		driver = guess_driver_can_write(dsn),
 		dataset_options = NULL, layer_options = NULL, quiet = FALSE, factorsAsCharacter = TRUE,
-		update = driver %in% db_drivers, delete_dsn = FALSE, delete_layer = FALSE) {
-	if (inherits(dsn, c("DBIObject", "PostgreSQLConnection"))) {
-		if (is.null(layer)) 
-			layer = deparse(substitute(obj))
-		return(dbWriteTable(dsn, name = layer, value = obj, ..., factorsAsCharacter = factorsAsCharacter))
-	}
-	if (length(list(...)))
-		stop(paste("unrecognized argument(s)", unlist(list(...)), "\n"))
+		update = driver %in% db_drivers, delete_dsn = FALSE, delete_layer = FALSE,
+		fid_column_name = NULL) {
+
 	if (missing(dsn))
 		stop("dsn should specify a data source or filename")
+	if (inherits(dsn, c("DBIObject", "PostgreSQLConnection", "Pool"))) {
+		if (inherits(dsn, "Pool")) {
+			if (! requireNamespace("pool", quietly = TRUE)) # nocov start
+				stop("package pool required, please install it first")
+			dsn = pool::poolCheckout(dsn)
+			on.exit(pool::poolReturn(dsn)) # nocov end
+		}
+		if (is.null(layer))
+			layer = deparse(substitute(obj))
+		dbWriteTable(dsn, name = layer, value = obj, ...,
+			factorsAsCharacter = factorsAsCharacter)
+		return(invisible(obj))
+	} else if (!inherits(dsn, "character")) { # add methods for other dsn classes here...
+		stop(paste("no st_write method available for dsn of class", class(dsn)[1]))
+	}
+
+	if (length(list(...)))
+		stop(paste("unrecognized argument(s)", names(list(...)), "\n"))
 	if (is.null(layer))
 		layer <- file_path_sans_ext(basename(dsn))
 
-	if (length(dsn) == 1 && file.exists(dsn))
-		dsn = enc2utf8(normalizePath(dsn))
+	if (length(dsn) == 1 && length(grep("~", dsn)) == 1) # resolve ~
+		dsn = normalizePath(dsn, mustWork = FALSE) # nocov
+
+	# this seems to be always a good idea:
+	dsn = enc2utf8(dsn)
 
 	geom = st_geometry(obj)
 	obj[[attr(obj, "sf_column")]] = NULL
@@ -319,23 +417,36 @@ st_write.sf = function(obj, dsn, layer = NULL, ...,
 		else
 			class(geom[[1]])[1]
 
+	fids = if (!is.null(fid_column_name)) {
+			fids = as.character(obj[[fid_column_name]])
+			obj[[fid_column_name]] = NULL
+			fids
+		} else
+			character(0)
+
 	ret = CPL_write_ogr(obj, dsn, layer, driver,
 		as.character(dataset_options), as.character(layer_options),
-		geom, dim, quiet, update, delete_dsn, delete_layer)
+		geom, dim, fids, quiet, update, delete_dsn, delete_layer)
 	if (ret == 1) { # try through temp file:
 		tmp = tempfile(fileext = paste0(".", tools::file_ext(dsn))) # nocov start
 		if (!quiet)
 			message(paste("writing first to temporary file", tmp))
 		if (CPL_write_ogr(obj, tmp, layer, driver,
 				as.character(dataset_options), as.character(layer_options),
-				geom, dim, quiet, update, delete_dsn, delete_layer) == 1)
+				geom, dim, fids, quiet, update, delete_dsn, delete_layer) == 1)
 			stop(paste("failed writing to temporary file", tmp))
 		if (!file.copy(tmp, dsn, overwrite = update || delete_dsn || delete_layer))
 			stop(paste("copying", tmp, "to", dsn, "failed"))
 		if (!file.remove(tmp))
 			warning(paste("removing", tmp, "failed"))
 	} # nocov end
-	invisible(NULL)
+	invisible(obj)
+}
+
+#' @name st_write
+#' @export
+st_write.data.frame <- function(obj, dsn, layer = NULL, ...) {
+    st_write.sf(obj = st_as_sf(obj), dsn = dsn, layer = layer, ...)
 }
 
 #' @name st_write
@@ -347,13 +458,17 @@ write_sf <- function(..., quiet = TRUE, delete_layer = TRUE) {
 #' Get GDAL drivers
 #'
 #' Get a list of the available GDAL drivers
-#' @param what character: "vector" or "raster", anything else will return all drivers.
-#' @details The drivers available will depend on the installation of GDAL/OGR, and can vary; the \code{st_drivers()}
-#' function shows which are available, and which may be written (but all are assumed to be readable). Note that stray
-#' files in data source directories (such as *.dbf) may lead to spurious errors that accompanying *.shp are missing.
-#' @return a \code{data.frame} with driver metadata
-#' @details field \code{vsi} refers to the driver's capability to read/create datasets through the VSI*L API.
+#' @param what character: `"vector"` or `"raster"`, anything else will return all
+#'   drivers.
+#' @details The drivers available will depend on the installation of GDAL/OGR,
+#'   and can vary; the `st_drivers()` function shows all the drivers that are
+#'   readable, and which may be written. The field `vsi` refers to the driver's
+#'   capability to read/create datasets through the VSI*L API. [See GDAL website
+#'   for additional details on driver
+#'   support](https://gdal.org/drivers/vector/index.html).
+#' @return A `data.frame` with driver metadata.
 #' @export
+#' @md
 #' @examples
 #' st_drivers()
 st_drivers = function(what = "vector") {
@@ -416,7 +531,7 @@ guess_driver = function(dsn) {
 
 	# find match: try extension first
 	drv = extension_map[tolower(tools::file_ext(dsn))]
-	if (any(grep(":", gsub(":[/\\]", "/", dsn))))
+	if (is_db_driver(dsn))
 		drv = prefix_map[tolower(strsplit(dsn, ":")[[1]][1])]
 
 	drv <- unlist(drv)
@@ -428,6 +543,10 @@ guess_driver = function(dsn) {
 	drv
 }
 
+is_db_driver = function(dsn) {
+	any(grep(":", gsub(":[/\\]", "/", dsn)))
+}
+
 guess_driver_can_write = function(dns, drv = guess_driver(dns)) {
   if(is.na(drv)) {
     stop("Could not guess driver for ", dns, call. = FALSE)
