@@ -3,13 +3,19 @@
 
 #include <ogrsf_frmts.h>
 
+#define RCPP_DEFAULT_INCLUDE_CALL false
 #include "Rcpp.h"
 
 #include "gdal_sf_pkg.h"
+#include "gdal_read.h"
 
-Rcpp::List allocate_out_list(OGRFeatureDefn *poFDefn, int n_features, bool int64_as_string) {
+Rcpp::List allocate_out_list(OGRFeatureDefn *poFDefn, int n_features, bool int64_as_string,
+		Rcpp::CharacterVector fid_column) {
 
-	int n = poFDefn->GetFieldCount() + poFDefn->GetGeomFieldCount();
+	if (fid_column.size() > 1)
+		Rcpp::stop("FID column name should be a length 1 character vector"); // #nocov
+
+	int n = poFDefn->GetFieldCount() + poFDefn->GetGeomFieldCount() + fid_column.size();
 	Rcpp::List out(n);
 	Rcpp::CharacterVector names(n);
 	for (int i = 0; i < poFDefn->GetFieldCount(); i++) {
@@ -58,6 +64,9 @@ Rcpp::List allocate_out_list(OGRFeatureDefn *poFDefn, int n_features, bool int64
 		names[i] = poFieldDefn->GetNameRef();
 	}
 
+	if (fid_column.size())
+		names[ poFDefn->GetFieldCount() ] = fid_column[0];
+
 	for (int i = 0; i < poFDefn->GetGeomFieldCount(); i++) {
 		// get the geometry fields:
 		OGRGeomFieldDefn *poGFDefn = poFDefn->GetGeomFieldDefn(i);
@@ -67,12 +76,12 @@ Rcpp::List allocate_out_list(OGRFeatureDefn *poFDefn, int n_features, bool int64
 		const char *geom_name = poGFDefn->GetNameRef();
 		if (*geom_name == '\0') {
 			if (i > 0)
-				names[i + poFDefn->GetFieldCount()] = geom + std::to_string(i); // c++11; #nocov
+				names[i + poFDefn->GetFieldCount() + fid_column.size()] = geom + std::to_string(i); // c++11; #nocov
 			else
-				names[i + poFDefn->GetFieldCount()] = geom;
+				names[i + poFDefn->GetFieldCount() + fid_column.size()] = geom;
 		} else
-			names[i + poFDefn->GetFieldCount()] = geom_name;
-		out[i + poFDefn->GetFieldCount()] = Rcpp::List(n_features); // ?
+			names[i + poFDefn->GetFieldCount() + fid_column.size()] = geom_name;
+		out[i + poFDefn->GetFieldCount() + fid_column.size()] = Rcpp::List(n_features); // ?
 	}
 
 	out.attr("names") = names;
@@ -81,7 +90,7 @@ Rcpp::List allocate_out_list(OGRFeatureDefn *poFDefn, int n_features, bool int64
 
 int to_multi_what(std::vector<OGRGeometry *> gv) {
 	bool points = false, multipoints = false,
-		lines = false, multilines = false, 
+		lines = false, multilines = false,
 		polygons = false, multipolygons = false;
 
 	for (unsigned int i = 0; i < gv.size(); i++) {
@@ -132,7 +141,7 @@ Rcpp::List CPL_get_layers(Rcpp::CharacterVector datasource, Rcpp::CharacterVecto
 		Rcpp::stop("argument datasource should have length 1.\n"); // #nocov
 	std::vector <char *> open_options = create_options(options, false);
 	GDALDataset *poDS;
-	poDS = (GDALDataset *) GDALOpenEx(datasource[0], GDAL_OF_VECTOR | GDAL_OF_READONLY, NULL, 
+	poDS = (GDALDataset *) GDALOpenEx(datasource[0], GDAL_OF_VECTOR | GDAL_OF_READONLY, NULL,
 		open_options.data(), NULL);
 	if (poDS == NULL) {
 		Rcpp::Rcout << "Cannot open data source " << datasource[0] << std::endl;
@@ -182,50 +191,8 @@ Rcpp::List CPL_get_layers(Rcpp::CharacterVector datasource, Rcpp::CharacterVecto
 	return out;
 }
 
-// [[Rcpp::export]]
-Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector layer, 
-		Rcpp::CharacterVector options, bool quiet, Rcpp::NumericVector toTypeUser,
-		bool promote_to_multi = true, bool int64_as_string = false) {
-	// adapted from the OGR tutorial @ www.gdal.org
-	std::vector <char *> open_options = create_options(options, quiet);
-	GDALDataset *poDS;
-	poDS = (GDALDataset *) GDALOpenEx( datasource[0], GDAL_OF_VECTOR | GDAL_OF_READONLY, NULL, 
-		open_options.data(), NULL );
-	if( poDS == NULL ) {
-		Rcpp::Rcout << "Cannot open data source " << datasource[0] << std::endl;
-		Rcpp::stop("Open failed.\n");
-	}
-
-	if (layer.size() == 0) { // no layer specified
-		switch (poDS->GetLayerCount()) {
-			case 0: { // error:
-				Rcpp::stop("No layers in datasource.");
-			}
-			case 1: { // silent:
-				OGRLayer *poLayer = poDS->GetLayer(0);
-				layer = Rcpp::CharacterVector::create(poLayer->GetName());
-				break;
-			}
-			default: { // select first layer: message + warning:
-				OGRLayer *poLayer = poDS->GetLayer(0);
-				layer = Rcpp::CharacterVector::create(poLayer->GetName());
-				if (! quiet) { // #nocov start
-					Rcpp::Rcout << "Multiple layers are present in data source " << datasource[0] << ", ";
-					Rcpp::Rcout << "reading layer `" << layer[0] << "'." << std::endl;
-					Rcpp::Rcout << "Use `st_layers' to list all layer names and their type in a data source." << std::endl;
-					Rcpp::Rcout << "Set the `layer' argument in `st_read' to read a particular layer." << std::endl;
-				} // #nocov end
-				Rcpp::Function warning("warning");
-				warning("automatically selected the first layer in a data source containing more than one.");
-			}
-		}
-	}
-
-	OGRLayer *poLayer = poDS->GetLayerByName(layer[0]);
-	if (poLayer == NULL) {
-		Rcpp::Rcout << "Cannot open layer " << layer[0] << std::endl;
-		Rcpp::stop("Opening layer failed.\n");
-	}
+Rcpp::List sf_from_ogrlayer(OGRLayer *poLayer, bool quiet, bool int64_as_string,
+		Rcpp::NumericVector toTypeUser, Rcpp::CharacterVector fid_column, bool promote_to_multi = true) {
 
 	double n_d = (double) poLayer->GetFeatureCount();
 	if (n_d > INT_MAX)
@@ -235,17 +202,14 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 	size_t n = (size_t) n_d; // what is List's max length?
 
 	std::vector<OGRFeature *> poFeatureV(n); // full archive
-
-	if (! quiet)
-		Rcpp::Rcout << "Reading layer `" << layer[0] << "' from data source `" << datasource[0] << // #nocov
-			"' using driver `" << poDS->GetDriverName() << "'" << std::endl;                       // #nocov
+	Rcpp::CharacterVector fids(n);
 
 	OGRFeatureDefn *poFDefn = poLayer->GetLayerDefn();
 
 	std::vector<OGRGeometry *> poGeometryV(n * poFDefn->GetGeomFieldCount());
 	// cycles column wise: 2nd el is 1st geometry, 2nd feature
 
-	Rcpp::List out = allocate_out_list(poFDefn, n, int64_as_string);
+	Rcpp::List out = allocate_out_list(poFDefn, n, int64_as_string, fid_column);
 
 	// read all features:
 	poLayer->ResetReading();
@@ -254,6 +218,8 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 	bool warn_int64 = false, has_null_geometries = false;
 	OGRFeature *poFeature;
 	while((poFeature = poLayer->GetNextFeature()) != NULL) {
+		// getFID:
+		fids[i] = std::to_string(poFeature->GetFID());
 
 		// feature attribute fields:
 		for (int iField = 0; iField < poFDefn->GetFieldCount(); iField++ ) {
@@ -313,9 +279,9 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 					poFeature->GetFieldAsDateTime(iField, &Year, &Month, &Day, &Hour, &Minute,
 						&Second, &TZFlag);
 					//  POSIXlt: sec   min  hour  mday   mon  year  wday  yday isdst ...
-					Rcpp::List dtlst = 
-						Rcpp::List::create((double) Second, (double) Minute, 
-						(double) Hour, (double) Day, (double) Month - 1, (double) Year - 1900, 
+					Rcpp::List dtlst =
+						Rcpp::List::create((double) Second, (double) Minute,
+						(double) Hour, (double) Day, (double) Month - 1, (double) Year - 1900,
 						0.0, 0.0, 0.0);
 					dtlst.attr("class") = "POSIXlt";
 					if (TZFlag == 100)
@@ -378,7 +344,7 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 					for (int j = 0; j < iv.size(); j++)
 						iv[j] = il[j];
 					lv[i] = iv;
-					} 
+					}
 					break;
 				case OFTInteger64List: {
 					Rcpp::List lv;
@@ -413,7 +379,7 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 					else
 						cv[i] = NA_STRING;
 					}
-					break;
+				break;
 			}
 		}
 
@@ -428,8 +394,13 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 		i++;
 	} // all read...
 
+	// add feature IDs if needed:
+	if (fid_column.size())
+		out[ poFDefn->GetFieldCount() ] = fids;
+
 	std::vector<OGRGeometry *> to_be_freed;
 	for (int iGeom = 0; iGeom < poFDefn->GetGeomFieldCount(); iGeom++ ) {
+
 		std::vector<OGRGeometry *> poGeom(n);
 		for (i = 0; i < n; i++)
 			poGeom[i] = poGeometryV[i + n * iGeom];
@@ -450,8 +421,8 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 				OGRGeometry *geom = poFeatureV[i]->StealGeometry(iGeom); // transfer ownership
 				if (geom == NULL)
 					geom = OGRGeometryFactory::createGeometry((OGRwkbGeometryType) toType); // #nocov
-				else if ((geom = 
-						OGRGeometryFactory::forceTo(geom, (OGRwkbGeometryType) toType, NULL)) 
+				else if ((geom =
+						OGRGeometryFactory::forceTo(geom, (OGRwkbGeometryType) toType, NULL))
 						== NULL)
 					Rcpp::stop("OGRGeometryFactory::forceTo returned NULL"); // #nocov
 				handle_error(poFeatureV[i]->SetGeomFieldDirectly(iGeom, geom));
@@ -487,11 +458,11 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 		Rcpp::List sfc = sfc_from_ogr(poGeom, false); // don't destroy
 		OGRGeomFieldDefn *fdfn = poFDefn->GetGeomFieldDefn(iGeom);
 		sfc.attr("crs") = get_crs(fdfn->GetSpatialRef()); // overwrite: see #449 for the reason why
-		out[iGeom + poFDefn->GetFieldCount()] = sfc;
+		out[iGeom + poFDefn->GetFieldCount() + fid_column.size()] = sfc;
 	}
 
 	if (warn_int64)
-		Rcpp::Rcout << "Integer64 values larger than " << dbl_max_int64 << 
+		Rcpp::Rcout << "Integer64 values larger than " << dbl_max_int64 <<
 			" lost significance after conversion to double;" << std::endl <<
 			"use argument int64_as_string = TRUE to import them lossless, as character" << std::endl;
 
@@ -500,7 +471,87 @@ Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector
 		OGRFeature::DestroyFeature(poFeatureV[i]);
 	for (i = 0; i < to_be_freed.size(); i++)
 		OGRGeometryFactory::destroyGeometry(to_be_freed[i]);
-	GDALClose(poDS);
 
 	return out;
 }
+
+// [[Rcpp::export]]
+Rcpp::List CPL_read_ogr(Rcpp::CharacterVector datasource, Rcpp::CharacterVector layer,
+		Rcpp::CharacterVector query,
+		Rcpp::CharacterVector options, bool quiet, Rcpp::NumericVector toTypeUser,
+		Rcpp::CharacterVector fid_column_name, Rcpp::CharacterVector drivers,
+		bool promote_to_multi = true, bool int64_as_string = false,
+		bool dsn_exists = true,
+		bool dsn_isdb = false) {
+
+	// adapted from the OGR tutorial @ www.gdal.org
+	std::vector <char *> open_options = create_options(options, quiet);
+	std::vector <char *> drivers_v = create_options(drivers, quiet);
+	GDALDataset *poDS;
+	poDS = (GDALDataset *) GDALOpenEx( datasource[0], GDAL_OF_VECTOR | GDAL_OF_READONLY, 
+		drivers.size() ? drivers_v.data() : NULL, open_options.data(), NULL );
+	if( poDS == NULL ) {
+		// could not open dsn
+		if( dsn_isdb ) {
+			Rcpp::stop("Cannot open %s; Check connection parameters.", datasource);
+		}
+		if( dsn_exists ) {
+			Rcpp::stop("Cannot open %s; %s %s",
+              datasource,
+              "The source could be corrupt or not supported.",
+              "See `st_drivers()` for a list of supported formats.");
+		}
+		Rcpp::stop("Cannot open %s; The file doesn't seem to exist.", datasource);
+	}
+
+	if (layer.size() == 0) { // no layer specified
+		switch (poDS->GetLayerCount()) {
+			case 0: { // error:
+				Rcpp::stop("No layers in datasource.");
+			}
+			case 1: { // silent:
+				OGRLayer *poLayer = poDS->GetLayer(0);
+				layer = Rcpp::CharacterVector::create(poLayer->GetName());
+				break;
+			}
+			default: { // select first layer: message + warning:
+				OGRLayer *poLayer = poDS->GetLayer(0);
+				layer = Rcpp::CharacterVector::create(poLayer->GetName());
+				if (! quiet) { // #nocov start
+					Rcpp::Rcout << "Multiple layers are present in data source " << datasource[0] << ", ";
+					Rcpp::Rcout << "reading layer `" << layer[0] << "'." << std::endl;
+					Rcpp::Rcout << "Use `st_layers' to list all layer names and their type in a data source." << std::endl;
+					Rcpp::Rcout << "Set the `layer' argument in `st_read' to read a particular layer." << std::endl;
+				} // #nocov end
+				Rcpp::Function warning("warning");
+				warning("automatically selected the first layer in a data source containing more than one.");
+			}
+		}
+	}
+
+	OGRLayer *poLayer;
+	if (! Rcpp::CharacterVector::is_na(query[0])) {
+		poLayer = poDS->ExecuteSQL(query[0], NULL, NULL);
+		if (poLayer == NULL)
+			Rcpp::stop("Query execution failed, cannot open layer.\n"); // #nocov
+	} else
+		poLayer = 	poDS->GetLayerByName(layer[0]);
+	if (poLayer == NULL) {
+		Rcpp::Rcout << "Cannot open layer " << layer[0] << std::endl;
+		Rcpp::stop("Opening layer failed.\n");
+	}
+
+	if (! quiet)
+		Rcpp::Rcout << "Reading layer `" << layer[0] << "' from data source `" << datasource[0] << // #nocov
+			"' using driver `" << poDS->GetDriverName() << "'" << std::endl;                       // #nocov
+
+	Rcpp::List out = sf_from_ogrlayer(poLayer, quiet, int64_as_string, toTypeUser, fid_column_name,
+		promote_to_multi);
+
+	// clean up if SQL was used https://www.gdal.org/classGDALDataset.html#ab2c2b105b8f76a279e6a53b9b4a182e0
+	if (! Rcpp::CharacterVector::is_na(query[0]))
+		poDS->ReleaseResultSet(poLayer);
+
+	GDALClose(poDS);
+	return out;
+}
