@@ -296,12 +296,12 @@ Rcpp::List CPL_geos_binop(Rcpp::List sfc0, Rcpp::List sfc1, std::string op, doub
 		}
 		out.attr("dim") = get_dim(sfc0.length(), sfc1.length());
 		ret_list = Rcpp::List::create(out);
-	} else if (op == "distance" || op == "Hausdorff" || op == "Frechet") { // return double matrix:
+	} else if (op == "Euclidean" || op == "distance" || op == "Hausdorff" || op == "Frechet") { // return double matrix:
 		// dist_fn, dist_parfn
 		Rcpp::NumericMatrix out(sfc0.length(), sfc1.length());
 		if (par <= 0.0) {
 			dist_fn dist_function;
-			if (op == "distance")
+			if (op == "Euclidean" || op == "distance")
 				dist_function = GEOSDistance_r;
 			else if (op == "Hausdorff")
 				dist_function = GEOSHausdorffDistance_r;
@@ -315,13 +315,20 @@ Rcpp::List CPL_geos_binop(Rcpp::List sfc0, Rcpp::List sfc1, std::string op, doub
 			}
 
 			for (size_t i = 0; i < gmv0.size(); i++) {
-				for (size_t j = 0; j < gmv1.size(); j++) {
-					double dist = -1.0;
-					if (dist_function(hGEOSCtxt, gmv0[i].get(), gmv1[j].get(), &dist) == 0) {
-						CPL_geos_finish(hGEOSCtxt); // #nocov
-						Rcpp::stop("GEOS error in GEOS_xx_Distance_r"); // #nocov
+				if (GEOSisEmpty_r(hGEOSCtxt, gmv0[i].get())) {
+					for (size_t j = 0; j < gmv1.size(); j++) // #nocov
+						out(i, j) = NA_REAL;                 // #nocov
+				} else for (size_t j = 0; j < gmv1.size(); j++) {
+					if (GEOSisEmpty_r(hGEOSCtxt, gmv1[j].get()))
+						out(i, j) = NA_REAL;
+					else {
+						double dist = -1.0;
+						if (dist_function(hGEOSCtxt, gmv0[i].get(), gmv1[j].get(), &dist) == 0) {
+							CPL_geos_finish(hGEOSCtxt); // #nocov
+							Rcpp::stop("GEOS error in GEOS_xx_Distance_r"); // #nocov
+						}
+						out(i, j) = dist;
 					}
-					out(i, j) = dist;
 				}
 				Rcpp::checkUserInterrupt();
 			}
@@ -339,13 +346,20 @@ Rcpp::List CPL_geos_binop(Rcpp::List sfc0, Rcpp::List sfc1, std::string op, doub
 			}
 
 			for (size_t i = 0; i < gmv0.size(); i++) {
-				for (size_t j = 0; j < gmv1.size(); j++) {
-					double dist = -1.0;
-					if (dist_function(hGEOSCtxt, gmv0[i].get(), gmv1[j].get(), par, &dist) == 0) {
-						CPL_geos_finish(hGEOSCtxt); // #nocov
-						Rcpp::stop("GEOS error in GEOS_xx_Distance_r"); // #nocov
+				if (GEOSisEmpty_r(hGEOSCtxt, gmv0[i].get())) {
+					for (size_t j = 0; j < gmv1.size(); j++)
+						out(i, j) = NA_REAL;
+				} else for (size_t j = 0; j < gmv1.size(); j++) {
+					if (GEOSisEmpty_r(hGEOSCtxt, gmv1[j].get()))
+						out(i, j) = NA_REAL;
+					else {
+						double dist = -1.0;
+						if (dist_function(hGEOSCtxt, gmv0[i].get(), gmv1[j].get(), par, &dist) == 0) {
+							CPL_geos_finish(hGEOSCtxt); // #nocov
+							Rcpp::stop("GEOS error in GEOS_xx_Distance_r"); // #nocov
+						}
+						out(i, j) = dist;
 					}
-					out(i, j) = dist;
 				}
 				Rcpp::checkUserInterrupt();
 			}
@@ -551,17 +565,36 @@ Rcpp::List CPL_geos_normalize(Rcpp::List sfc) { // #nocov start
 
 // [[Rcpp::export]]
 Rcpp::List CPL_geos_union(Rcpp::List sfc, bool by_feature = false) {
+
+	if (sfc.size() == 0)
+		return sfc; // #nocov
+
 	int dim = 2;
 	GEOSContextHandle_t hGEOSCtxt = CPL_geos_init();
 	std::vector<GeomPtr> gmv = geometries_from_sfc(hGEOSCtxt, sfc, &dim);
 	std::vector<GeomPtr> gmv_out(by_feature ? sfc.size() : 1);
+
 	if (by_feature) {
 		for (int i = 0; i < sfc.size(); i++) {
 			gmv_out[i] = geos_ptr(GEOSUnaryUnion_r(hGEOSCtxt, gmv[i].get()), hGEOSCtxt);
 		}
 	} else {
-		GeomPtr gc = geos_ptr(GEOSGeom_createCollection_r(hGEOSCtxt, GEOS_GEOMETRYCOLLECTION, to_raw(gmv).data(), gmv.size()), hGEOSCtxt);
-		gmv_out[0] = geos_ptr(GEOSUnaryUnion_r(hGEOSCtxt, gc.get()), hGEOSCtxt);
+		bool all_inputs_same = true;
+
+		// check to see if all geometries are identical, as in a call to summarize(..., do_union=TRUE)
+		for (size_t i = 1; i < gmv.size(); i++) {
+			if (!GEOSEqualsExact_r(hGEOSCtxt, gmv[0].get(), gmv[i].get(), 0.0)) {
+				all_inputs_same = false;
+				break;
+			}
+		}
+
+		if (all_inputs_same) {
+			gmv_out[0] = std::move(gmv[0]);
+		} else {
+			GeomPtr gc = geos_ptr(GEOSGeom_createCollection_r(hGEOSCtxt, GEOS_GEOMETRYCOLLECTION, to_raw(gmv).data(), gmv.size()), hGEOSCtxt);
+			gmv_out[0] = geos_ptr(GEOSUnaryUnion_r(hGEOSCtxt, gc.get()), hGEOSCtxt);
+		}
 	}
 
 	Rcpp::List out(sfc_from_geometry(hGEOSCtxt, gmv_out, dim));
@@ -784,10 +817,21 @@ Rcpp::List CPL_geos_op2(std::string op, Rcpp::List sfcx, Rcpp::List sfcy) {
 	m(_, 0) = Rcpp::NumericVector(index_x.begin(), index_x.end());
 	m(_, 1) = Rcpp::NumericVector(index_y.begin(), index_y.end());
 
-	Rcpp::List ret(sfc_from_geometry(hGEOSCtxt, out, dim));
+	Rcpp::List ret;
+	if ((x.size() == 0 || y.size() == 0) && op != "intersection") {
+		if (op == "union" || op == "sym_difference") { // return "the other"
+			if (y.size() == 0)
+				ret = sfcx;
+			else
+				ret = sfcy;
+		} else // "difference" is asymmetric: x - 0 -> return x
+			ret = sfcx;
+	} else {
+		ret = sfc_from_geometry(hGEOSCtxt, out, dim);
+		ret.attr("crs") = sfcx.attr("crs");
+		ret.attr("idx") = m;
+	}
 	CPL_geos_finish(hGEOSCtxt);
-	ret.attr("crs") = sfcx.attr("crs");
-	ret.attr("idx") = m;
 	return ret;
 }
 
@@ -956,8 +1000,8 @@ Rcpp::List CPL_nary_difference(Rcpp::List sfc) {
 					contained = chk_(GEOSContains_r(hGEOSCtxt, out[tree_sel[j]].get(), geom.get()));
 					if (contained)
 						break;
-					// test if the items overlap with geom
-					if (chk_(GEOSOverlaps_r(hGEOSCtxt, geom.get(), out[tree_sel[j]].get()))) {
+					// test if the items intersect with geom
+					if (chk_(GEOSIntersects_r(hGEOSCtxt, geom.get(), out[tree_sel[j]].get()))) {
 						// if they do then erase overlapping parts from geom
 						geom = geos_ptr(GEOSDifference_r(hGEOSCtxt, geom.get(), out[tree_sel[j]].get()), hGEOSCtxt);
 						if (geom == nullptr)
